#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <limits.h>
#include <sys/stat.h>
#include <regex.h>

#include "config.h"
#include "imapfilter.h"
#include "pathnames.h"


extern char logfile[PATH_MAX];
extern unsigned int options;
extern char charset[CHARSET_LEN];
extern unsigned int flags;
extern unsigned int interval;
extern long timeout;
extern char *home;

#ifdef ENCRYPTED_PASSWORDS
char *passphr = NULL;		/* Master password to access the passwords
				 * file. */
#endif


int parse_config(FILE * fd);
void set_options(char *line, regmatch_t * match);

int parse_passwords(FILE * fd);


/*
 * Find the path to configuration file, open it and call parse_config().
 */
int
read_config(char *cfg)
{
	int r;
	FILE *fd;
	char *c;

	c = NULL;

	if (cfg == NULL) {
		cfg = c = (char *)xmalloc(PATH_MAX * sizeof(char));
		snprintf(cfg, PATH_MAX, "%s/%s", home, PATHNAME_CONFIG_FILE);
	}
#ifdef DEBUG
	fprintf(stderr, "debug: configuration file: '%s'\n", cfg);
#endif
#ifdef CHECK_PERMISSIONS
	check_file_perms(cfg, S_IRUSR | S_IWUSR);
#endif
	fd = fopen(cfg, "r");
	if (fd == NULL)
		fatal(ERROR_FILE_OPEN, "opening config file %s; %s\n", cfg,
		    strerror(errno));

	if (c != NULL)
		xfree(c);

	if ((r = parse_config(fd)))
		fatal(ERROR_CONFIG_PARSE,
		    "parse error in config file at row %d\n", r);

	fclose(fd);

#ifdef DEBUG
	fprintf(stderr, "debug: options: %0#10x '%s'\n", options, charset);
#endif

	return 0;
}


/*
 * Parse configuration file.
 */
int
parse_config(FILE * fd)
{
	int i, r;
	unsigned int row;
	char line[LINE_MAX];
	regex_t creg[13];
	regmatch_t match[11];
	const char *reg[13] = {
		"^([[:blank:]]*\n|#.*\n)$",

		"^[[:blank:]]*ACCOUNT[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]+"
		"(([[:graph:]]+):([[:graph:]]*)|([[:graph:]]+))@"
		"([[:alnum:].-]+)(:[[:digit:]]+)?[[:blank:]]*"
		"([[:blank:]]SSL2|[[:blank:]]SSL3|[[:blank:]]TLS1)?"
		"[[:blank:]]*\n$",

		"^[[:blank:]]*FOLDER[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]+"
		"([[:print:]]+)[[:blank:]]*\n$",

		"^[[:blank:]]*FILTER[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]*"
		"([[:blank:]]OR|[[:blank:]]AND)?[[:blank:]]*\n$",

		"^[[:blank:]]*ACTION[[:blank:]]+(DELETE|"
		"COPY[[:blank:]]+(\"[[:print:]]*\"|[[:graph:]]+)|"
		"MOVE[[:blank:]]+(\"[[:print:]]*\"|[[:graph:]]+)|"
		"RCOPY[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]+"
		"(\"[[:print:]]*\"|[[:graph:]]+)|"
		"RMOVE[[:blank:]]+([[:alnum:]_-]+)[[:blank:]]+"
		"(\"[[:print:]]*\"|[[:graph:]]+)|"
		"FLAG[[:blank:]]+(REPLACE|ADD|REMOVE)[[:blank:]]+"
		"([[:alpha:],]+)|"
		"LIST)[[:blank:]]*([[:graph:]]*)[[:blank:]]*\n$",

		"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|"
		"AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*"
		"(ANSWERED|DELETED|DRAFT|FLAGGED|NEW|OLD|RECENT|SEEN|"
		"UNANSWERED|UNDELETED|UNDRAFT|UNFLAGGED|UNSEEN)[[:blank:]]*\n$",

		"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|"
		"AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*"
		"(BCC|BODY|CC|FROM|SUBJECT|TEXT|TO)[[:blank:]]+"
		"(\"[[:print:]]*\"|[[:graph:]]+)[[:blank:]]*\n$",

		"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|"
		"AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*"
		"(HEADER)[[:blank:]]+(\"[[:print:]]*\"|[[:graph:]]+)"
		"[[:blank:]]+(\"[[:print:]]*\"|[[:graph:]]+)[[:blank:]]*\n$",

		"^[[:blank:]]*(MASK[[:blank:]])?[[:blank:]]*(OR[[:blank:]]|"
		"AND[[:blank:]])?[[:blank:]]*(NOT[[:blank:]])?[[:blank:]]*"
		"(LARGER|SMALLER|OLDER|NEWER)[[:blank:]]+([[:digit:]]+)"
		"[[:blank:]]*\n$",

		"^[[:blank:]]*JOB[[:blank:]]+([[:alnum:],_-]+)[[:blank:]]+"
		"([[:alnum:],_-]+)[[:blank:]]*\n$",

		"^[[:blank:]]*(SET[[:blank:]])?[[:blank:]]*(LOGFILE|CHARSET)"
		"[[:blank:]]*=[[:blank:]]*(\"[[:print:]]*\"|[[:graph:]]+)"
		"[[:blank:]]*\n$",

		"^[[:blank:]]*(SET[[:blank:]])?[[:blank:]]*(ERRORS|EXPUNGE|"
		"HEADERS|NAMESPACE|SUBSCRIBE)[[:blank:]]*=[[:blank:]]*"
		"(YES|NO)[[:blank:]]*\n$",

		"^[[:blank:]]*(SET[[:blank:]])?[[:blank:]]*(DAEMON|TIMEOUT)"
		"[[:blank:]]*=[[:blank:]]*([[:digit:]]+)\n$"
	};

	r = row = 0;

	for (i = 0; i < 13; i++)
		regcomp(&creg[i], reg[i], REG_EXTENDED | REG_ICASE);

	/* First process all the variables.  This is done because some
	 * variables' arguments are used before the other commands are
	 * processed. */
	while (fgets(line, LINE_MAX - 1, fd))
		if (!regexec(&creg[10], line, 4, match, 0) ||
		    !regexec(&creg[11], line, 4, match, 0) ||
		    !regexec(&creg[12], line, 4, match, 0))
			set_options(line, match);

	/* Then rewind and process everything else. */
	fseek(fd, 0L, SEEK_SET);
	while (fgets(line, LINE_MAX - 1, fd)) {
		row++;
		if (!regexec(&creg[0], line, 0, match, 0))
			continue;
		else if (!regexec(&creg[1], line, 9, match, 0))
			set_account(line, match);
		else if (!regexec(&creg[2], line, 3, match, 0))
			r = set_mboxgrp(line, match);
		else if (!regexec(&creg[3], line, 3, match, 0))
			r = set_filter(line, match);
		else if (!regexec(&creg[4], line, 11, match, 0))
			r = set_action(line, match);
		else if (!regexec(&creg[5], line, 5, match, 0))
			r = set_mask(line, match, MASK_MATCH_1);
		else if (!regexec(&creg[6], line, 6, match, 0))
			r = set_mask(line, match, MASK_MATCH_2);
		else if (!regexec(&creg[7], line, 7, match, 0))
			r = set_mask(line, match, MASK_MATCH_3);
		else if (!regexec(&creg[8], line, 6, match, 0))
			r = set_mask(line, match, MASK_MATCH_4);
		else if (!regexec(&creg[9], line, 3, match, 0))
			r = set_job(line, match);
		/* Skip variable processing. */
		else if (!regexec(&creg[10], line, 4, match, 0) ||
			    !regexec(&creg[11], line, 4, match, 0) ||
		    !regexec(&creg[12], line, 4, match, 0));
		else
			return row;

		if (r == ERROR_CONFIG_PARSE)
			return row;
	}

	for (i = 0; i < 13; i++)
		regfree(&creg[i]);

	destroy_unneeded();

	return 0;
}


/*
 * Signal SIGUSR1 received, destroy all data structures and reread the
 * configuration file.
 */
void
reread_config(char *cfg)
{
	destroy_all();
	read_config(cfg);
#ifdef ENCRYPTED_PASSWORDS
	read_passwords();
#endif
	flags &= ~(FLAG_SIGUSR1_RECEIVED);
}


/*
 * Set other options found in config file.
 */
void
set_options(char *line, regmatch_t * m)
{
	if (!strncasecmp(line + m[2].rm_so, "logfile", 7)) {
		if (*logfile == '\0') {
			if (*(line + m[3].rm_so) == '"' &&
			    *(line + m[3].rm_eo - 1) == '"')
				strncat(logfile, line + m[3].rm_so + 1,
				    min((m[3].rm_eo - m[3].rm_so - 2),
					PATH_MAX - 1));
			else
				strncat(logfile, line + m[3].rm_so,
				    min((m[3].rm_eo - m[3].rm_so),
					PATH_MAX - 1));
		}
	} else if (!strncasecmp(line + m[2].rm_so, "charset", 7)) {
		if (*(line + m[3].rm_so) == '"' &&
		    *(line + m[3].rm_eo - 1) == '"')
			strncat(charset, line + m[3].rm_so + 1,
			    min((m[3].rm_eo - m[3].rm_so - 2),
				CHARSET_LEN - 1));
		else
			strncat(charset, line + m[3].rm_so,
			    min((m[3].rm_eo - m[3].rm_so), CHARSET_LEN - 1));
	} else if (!strncasecmp(line + m[2].rm_so, "errors", 6)) {
		if (!strncasecmp(line + m[3].rm_so, "yes", 3))
			options |= OPTION_ERRORS;
		else
			options &= ~(OPTION_ERRORS);
	} else if (!strncasecmp(line + m[2].rm_so, "expunge", 7)) {
		if (!strncasecmp(line + m[3].rm_so, "yes", 3))
			options |= OPTION_EXPUNGE;
		else
			options &= ~(OPTION_EXPUNGE);
	} else if (!strncasecmp(line + m[2].rm_so, "header", 6)) {
		if (!strncasecmp(line + m[3].rm_so, "yes", 3))
			options |= OPTION_HEADERS;
		else
			options &= ~(OPTION_HEADERS);
	} else if (!strncasecmp(line + m[2].rm_so, "namespace", 9)) {
		if (!strncasecmp(line + m[3].rm_so, "yes", 3))
			options |= OPTION_NAMESPACE;
		else
			options &= ~(OPTION_NAMESPACE);
	} else if (!strncasecmp(line + m[2].rm_so, "subscribe", 9)) {
		if (!strncasecmp(line + m[3].rm_so, "yes", 3))
			options |= OPTION_SUBSCRIBE;
		else
			options &= ~(OPTION_SUBSCRIBE);
	} else if (!strncasecmp(line + m[2].rm_so, "timeout", 7)) {
		errno = 0;
		timeout = strtol(line + m[3].rm_so, NULL, 10);
		if (errno)
			timeout = 0;
	} else if (!strncasecmp(line + m[2].rm_so, "daemon", 6) &&
	    !(options & OPTION_DAEMON_MODE)) {
		options |= OPTION_DAEMON_MODE;
		errno = 0;
		interval = strtoul(line + m[3].rm_so, NULL, 10);
		if (errno)
			interval = 0;
	}
}


#ifdef ENCRYPTED_PASSWORDS
/*
 * Open password file and call parse_passwords().
 */
int
read_passwords(void)
{
	FILE *fd;
	char pwfile[PATH_MAX];

	if (!(flags & FLAG_BLANK_PASSWORD))
		return ERROR_CONFIG_PARSE;

	if (!passphr) {
		passphr = (char *)smalloc(PASSPHRASE_LEN);
		*passphr = '\0';
	}
	snprintf(pwfile, PATH_MAX, "%s/%s", home, PATHNAME_PASSWORD_FILE);
#ifdef DEBUG
	fprintf(stderr, "debug: passwords file: '%s'\n", pwfile);
#endif

	if (!exists_file(pwfile))
		return ERROR_FILE_OPEN;

#ifdef CHECK_PERMISSIONS
	check_file_perms(pwfile, S_IRUSR | S_IWUSR);
#endif

	fd = fopen(pwfile, "r");
	if (fd == NULL)
		fatal(ERROR_FILE_OPEN, "opening passwords file %s; %s\n",
		    pwfile, strerror(errno));

	parse_passwords(fd);

	fclose(fd);

	return 0;
}


/*
 *  Parse unencrypted password file.
 */
int
parse_passwords(FILE * fd)
{
	int t;
	char *pe;
	char user[USERNAME_LEN], serv[SERVER_LEN];
	unsigned char *buf;
	char *c, *cp, *line;
	regex_t creg;
	regmatch_t match[4];
	const char *reg;

	t = 3;
	pe = NULL;
	reg = "([[:alnum:].-]+) ([[:graph:]]+) ([[:graph:]]+)";

	if (!(flags & FLAG_DAEMON_MODE)) {
		do {
			fseek(fd, 0L, SEEK_SET);
			printf("Enter master passphrase: ");
			get_password(passphr, PASSPHRASE_LEN);
		} while (decrypt_passwords(&buf, fd) && --t != 0);

		if (!t)
			return ERROR_PASSPHRASE;
	} else if (decrypt_passwords(&buf, fd))
		fatal(ERROR_DECRYPT,
		    "failed decrypting passwords in daemon mode\n");

	c = cp = sstrdup(buf);

	regcomp(&creg, reg, REG_EXTENDED | REG_ICASE);

	line = strtok_r(c, "\n", &c);
	while (line != NULL && !regexec(&creg, line, 4, match, 0)) {
		user[0] = serv[0] = '\0';

		strncat(serv, line + match[1].rm_so,
		    min(match[1].rm_eo - match[1].rm_so, SERVER_LEN - 1));
		strncat(user, line + match[2].rm_so,
		    min(match[2].rm_eo - match[2].rm_so, USERNAME_LEN - 1));

		if ((pe = (char *)find_password(user, serv)))
			strncat(pe, line + match[3].rm_so,
			    min(match[3].rm_eo - match[3].rm_so,
				PASSWORD_LEN - 1));

		line = strtok_r(NULL, "\n", &c);
	}

	regfree(&creg);
	sfree(cp);
	sfree(buf);

	return 0;
}
#endif				/* ENCRYPTED_PASSWORDS */
