#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <fcntl.h>

#include "config.h"
#include "imapfilter.h"

#ifdef SSL_TLS
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif


extern unsigned int options;
extern conn_t connpri, connaux;

long timeout = -1;		/* Server non-response timeout in seconds. */


/*
 * Connect to mail server.
 */
int
init_connection(conn_t * conn, char *serv, unsigned short int port,
    unsigned int protocol)
{
	struct sockaddr_in sa;
	struct hostent *he;

#ifndef SSL_TLS
	if (protocol != SSL_DISABLED) {
		error("SSL not supported by this build\n");
		return ERROR_SSL;
	}
#endif

	memset((char *)&sa, 0, sizeof(struct sockaddr_in));

	conn->sock = socket(AF_INET, SOCK_STREAM, 0);

	if (conn->sock < 0) {
		error("create socket; %s\n", strerror(errno));
		return ERROR_NETWORK;
	}
	if (!(he = gethostbyname(serv))) {
		error("get network host entry of %s; %s\n", serv,
		    strerror(errno));
		close_connection(conn);
		return ERROR_NETWORK;
	}
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sa.sin_addr = *(struct in_addr *) he->h_addr;

	if (connect(conn->sock, (struct sockaddr *) & sa,
		sizeof(struct sockaddr))) {
		error("initiating connection to %s; %s\n", serv,
		    strerror(errno));
		close_connection(conn);
		return ERROR_NETWORK;
	}
	if (conn == &connpri)
		info("Connected to %s.\n", he->h_name);

#ifdef SSL_TLS
	if (protocol != SSL_DISABLED)
		if (!init_secure_connection(conn, protocol))
			return 0;
		else
			return ERROR_SSL;
	else
		conn->ssl = NULL;
#endif

	return 0;
}


#ifdef SSL_TLS
/*
 * Initialize Secure Socket Layer connection.
 */
int
init_secure_connection(conn_t * conn, unsigned int protocol)
{
	int e;
	SSL_CTX *ctx;
	SSL_METHOD *method;

	method = NULL;

	SSL_library_init();
	SSL_load_error_strings();

	switch (protocol) {
	case SSL_TLS_V1:
		method = TLSv1_client_method();
		break;
	case SSL_SSL_V2:
	case SSL_SSL_V3:
		method = SSLv23_client_method();
		break;
	}

	if (!(ctx = SSL_CTX_new(method)))
		goto fail;

	if (!(conn->ssl = SSL_new(ctx)))
		goto fail;

	SSL_set_fd(conn->ssl, conn->sock);

	if ((e = SSL_connect(conn->ssl)) <= 0) {
		SSL_get_error(conn->ssl, e);
		error("initiating SSL connection; %s\n",
		    ERR_error_string(ERR_get_error(), NULL));
		goto fail;
	}
	if (options & OPTION_DETAILS_VERBOSE) {
		SSL_CIPHER *cipher;
		char *ver;
		const char *name;
		int usebits, algbits;

		cipher = SSL_get_current_cipher(conn->ssl);
		ver = SSL_CIPHER_get_version(cipher);
		name = SSL_CIPHER_get_name(cipher);
		usebits = SSL_CIPHER_get_bits(cipher, &algbits);
		printf("SSL/TLS handshake completed: %s with cipher %s "
		    "(%d/%d bits).\n", ver, name, usebits, algbits);
	}
	if (get_cert(conn))
		goto fail;

	SSL_CTX_free(ctx);

	return 0;

fail:
	conn->ssl = NULL;
	SSL_CTX_free(ctx);

	return ERROR_SSL;
}
#endif				/* SSL_TLS */


/*
 * Disconnect from mail server.
 */
int
close_connection(conn_t * conn)
{
#ifdef SSL_TLS
	if (conn->ssl) {
		SSL_shutdown(conn->ssl);
		SSL_free(conn->ssl);
		conn->ssl = NULL;
	}
#endif
	if (conn->sock >= 0 && close(conn->sock) == -1) {
		error("closing socket; %s\n", strerror(errno));
		return ERROR_NETWORK;
	} else {
		conn->sock = -1;
		return 0;
	}
}


/*
 * Read data from socket.
 */
int
socket_read(conn_t * conn, char *buf)
{
	int f, e, s;
	fd_set fds;
	struct timeval tv;
	struct timeval *tvp;

	e = 0;
	s = 1;
	tvp = NULL;

	memset(buf, 0, RESPONSE_BUF + 1);

	if (timeout >= 0) {
		tv.tv_sec = timeout;
		tv.tv_usec = 0;
		tvp = &tv;
	}
	f = fcntl(conn->sock, F_GETFL, 0);
	fcntl(conn->sock, F_SETFL, f | O_NONBLOCK);

	FD_ZERO(&fds);
	FD_SET(conn->sock, &fds);

#ifdef SSL_TLS
	if (conn->ssl) {
		while (SSL_pending(conn->ssl) > 0 ||
		    ((s = select(conn->sock + 1, &fds, NULL, NULL, tvp)) > 0 &&
			FD_ISSET(conn->sock, &fds))) {
			e = SSL_read(conn->ssl, buf, RESPONSE_BUF);

			if (e > 0)
				break;

			switch (SSL_get_error(conn->ssl, e)) {
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_WRITE:
				continue;
			case SSL_ERROR_ZERO_RETURN:
				return ERROR_NETWORK;
			case SSL_ERROR_SYSCALL:
			case SSL_ERROR_SSL:
				fatal(ERROR_NETWORK, "reading data; %s\n",
				    ERR_error_string(ERR_get_error(), NULL));
			default:
				fatal(ERROR_NETWORK,
				    "undefined ssl error while reading data\n");
			}
		}
	} else
#endif
	{
		if ((s = select(conn->sock + 1, &fds, NULL, NULL, tvp)) > 0 &&
		    FD_ISSET(conn->sock, &fds))
			e = read(conn->sock, buf, RESPONSE_BUF);

		if (e == -1)
			fatal(ERROR_NETWORK, "reading data; %s",
			    strerror(errno));
		else if (e == 0)
			return ERROR_NETWORK;
	}

	fcntl(conn->sock, F_SETFL, f);

	if (s == -1)
		fatal(ERROR_NETWORK, "waiting to read from socket; %s\n",
		    strerror(errno));
	else if (s == 0)
		fatal(ERROR_NETWORK,
		    "timeout period expired while waiting to read "
		    "data\n");

	return 0;
}


/*
 * Write data to socket.
 */
int
socket_write(conn_t * conn, char *data)
{
	int f, e, s;
	fd_set fds;
	struct timeval tv;
	struct timeval *tvp = NULL;

	e = 0;
	s = 1;

	if (timeout >= 0) {
		tv.tv_sec = timeout;
		tv.tv_usec = 0;
		tvp = &tv;
	}
	f = fcntl(conn->sock, F_GETFL, 0);
	fcntl(conn->sock, F_SETFL, f | O_NONBLOCK);

	FD_ZERO(&fds);
	FD_SET(conn->sock, &fds);

#ifdef SSL_TLS
	if (conn->ssl) {
		while ((s = select(conn->sock + 1, NULL, &fds, NULL, tvp) > 0 &&
			FD_ISSET(conn->sock, &fds))) {
			e = SSL_write(conn->ssl, data, strlen(data));

			if (e > 0)
				break;

			switch (SSL_get_error(conn->ssl, e)) {
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_WRITE:
				continue;
			case SSL_ERROR_ZERO_RETURN:
				return ERROR_NETWORK;
			case SSL_ERROR_SYSCALL:
			case SSL_ERROR_SSL:
				fatal(ERROR_NETWORK, "writing data; %s\n",
				    ERR_error_string(ERR_get_error(), NULL));
			default:
				fatal(ERROR_NETWORK,
				    "undefined ssl error while writing data\n");
			}
		}
	} else
#endif
	{
		if ((s = select(conn->sock + 1, NULL, &fds, NULL, tvp)) > 0 &&
		    FD_ISSET(conn->sock, &fds))
			e = write(conn->sock, data, strlen(data));

		if (e == -1)
			fatal(ERROR_NETWORK, "writing data; %s",
			    strerror(errno));
	}

	fcntl(conn->sock, F_SETFL, f);

	if (s == -1)
		fatal(ERROR_NETWORK, "waiting to write to socket; %s\n",
		    strerror(errno));
	else if (s == 0)
		fatal(ERROR_NETWORK,
		    "timeout period expired while waiting to write "
		    "data\n");

	return 0;
}
